@@ -1,18 +1,15 @@
 package Inline::Java::Portable ;
 @Inline::Java::Portable::ISA = qw(Exporter) ;
 
-@EXPORT = qw(portable) ;
-
+@EXPORT = qw(portable make_classpath get_server_jar get_user_jar) ;
 
 use strict ;
-
-$Inline::Java::Portable::VERSION = '0.31' ;
-
-
 use Exporter ;
 use Carp ;
 use Config ;
+use Cwd ;
 use File::Find ;
+use File::Spec ;
 
 # Here is some code to figure out if we are running on command.com
 # shell under Windows.
@@ -30,7 +27,6 @@ my $COMMAND_COM =
 	) || 0 ;
 
 
-
 sub debug {
 	if (Inline::Java->can("debug")){
 		return Inline::Java::debug(@_) ;
@@ -38,51 +34,90 @@ sub debug {
 }
 
 
-# Here in Inline <= 0.43 there is a portability issue
-# with the mkpath function. It splits directly on '/'.
-# We assume this will be fixed in 0.44
-sub mkpath {
-	my $o = shift ;
-	my $path = shift ;
+# Cleans the CLASSPATH environment variable and adds
+# the paths specified.
+sub make_classpath {
+	my @paths = @_ ;
 
-	if ($Inline::VERSION <= 0.43){
-		my $sep = File::Spec->catdir('', '') ;
-		$sep = quotemeta($sep) ;
-		$path =~ s/$sep/\//g ;
+	my @list = () ;
+	if (defined($ENV{CLASSPATH})){
+		push @list, $ENV{CLASSPATH} ;
 	}
-	
-	return $o->Inline::mkpath($path) ;
-} ;
+	push @list, @paths ;
 
+	my $sep = portable("ENV_VAR_PATH_SEP_CP") ;
+	my @cp = split(/$sep+/, join($sep, @list)) ;
 
-# Here in Inline <= 0.43 there is a portability issue
-# with the rmpath function. It splits directly on '/'.
-# We assume this will be fixed in 0.44
-sub rmpath {
-	my $o = shift ;
-	my $prefix = shift ;
-	my $path = shift ;
-	
-	if ($Inline::VERSION <= 0.43){
-		my $sep = File::Spec->catdir('', '') ;
-		$sep = quotemeta($sep) ;
-		$path =~ s/$sep/\//g ;
+	# Clean up paths
+	foreach my $p (@cp){
+		$p =~ s/^\s+// ;
+		$p =~ s/\s+$// ;
 	}
-	
-	return $o->Inline::rmpath($prefix, $path) ;
-} ;
+
+	# Remove duplicates, remove invalids but preserve order
+	my @fcp = () ;
+	my %cp = map {$_ => 1} @cp ;
+	foreach my $p (@cp){
+		if (($p)&&($cp{$p})&&(-e $p)){
+			my $fp = (-d $p ? Cwd::abs_path($p) : $p) ;
+			push @fcp, portable("SUB_FIX_CLASSPATH", $fp) ;
+			delete $cp{$p} ;
+		}
+		else{
+			Inline::Java::debug(2, "classpath candidate '$p' scraped") ;
+		}
+	}
+
+	my $cp = join($sep, @fcp) ;
+
+	return (wantarray ? @fcp : $cp) ;
+}
 
 
+sub get_jar_dir {
+	my $path = $INC{"Inline/Java.pm"} ;
+	my ($v, $d, $f) = File::Spec->splitpath($path) ;
+
+	# This undef for the file should be ok.
+	my $dir = File::Spec->catpath($v, $d, 'Java', undef) ;
+
+	return Cwd::abs_path($dir) ;
+}
+
+
+sub get_server_jar {
+	return File::Spec->catfile(get_jar_dir(), 'InlineJavaServer.jar') ;
+}
+
+
+sub get_user_jar {
+	return File::Spec->catfile(get_jar_dir(), 'InlineJavaUser.jar') ;
+}
+
+
+# This maybe could be made more stable
 sub find_classes_in_dir {
 	my $dir = shift ;
 
 	my @ret = () ;
 	find(sub {
-		my $file = $_ ;
-		if ($file =~ /\.class$/){
-			push @ret, $file ;
+		my $f = $_ ;
+		if ($f =~ /\.class$/){
+			my $file = $File::Find::name ;
+			my $fdir = $File::Find::dir ;
+			my @dirs = File::Spec->splitdir($fdir) ;
+			# Remove '.'
+			shift @dirs ;
+			# Add an empty dir to get the last '.' (for '.class')
+			if ((! scalar(@dirs))||($dirs[-1] ne '')){
+				push @dirs, '' ;
+			}
+			my $pkg = (scalar(@dirs) ? join('.', @dirs) : '') ;
+			my $class = "$pkg$f" ;
+			$class =~ s/\.class$// ;
+			push @ret, {file => $file, class => $class} ;
 		}
-	}, $dir) ;	
+	}, $dir) ;
 
 	return @ret ;
 }
@@ -96,6 +131,7 @@ sub portable {
 		EXE_EXTENSION		=>	$Config{exe_ext},
 		GOT_ALARM			=>  $Config{d_alarm} || 0,
 		GOT_FORK			=>	$Config{d_fork} || 0,
+		GOT_NEXT_FREE_PORT	=>	1,
 		ENV_VAR_PATH_SEP	=>	$Config{path_sep},
 		SO_EXT				=>	$Config{dlext},
 		PREFIX				=>	$Config{prefix},
@@ -104,9 +140,12 @@ sub portable {
 		SO_LIB_PATH_VAR		=>	'LD_LIBRARY_PATH',
 		ENV_VAR_PATH_SEP_CP	=>	':',
 		IO_REDIR			=>  '2>&1',
+		MAKE				=>	'make',
 		DEV_NULL			=>  '/dev/null',
 		COMMAND_COM			=>  0,
 		SUB_FIX_CLASSPATH	=>	undef,
+		SUB_FIX_CMD_QUOTES	=>	undef,
+		SUB_FIX_MAKE_QUOTES	=>	undef,
 		JVM_LIB				=>	'libjvm.so',
 		JVM_SO				=>	'libjvm.so',
 	} ;
@@ -116,12 +155,24 @@ sub portable {
 			ENV_VAR_PATH_SEP_CP	=>	';',
 			# 2>&1 doesn't work under command.com
 			IO_REDIR			=>  ($COMMAND_COM ? '' : undef),
+			MAKE				=>	'nmake',
 			DEV_NULL			=>  'nul',
 			COMMAND_COM			=>	$COMMAND_COM,
 			SO_LIB_PATH_VAR		=>	'PATH',
 			DETACH_OK			=>	0,
 			JVM_LIB				=>	'jvm.lib',
 			JVM_SO				=>	'jvm.dll',
+			GOT_NEXT_FREE_PORT	=>	0,
+			SUB_FIX_CMD_QUOTES	=>	($COMMAND_COM ? undef : sub {
+				my $val = shift ;
+				$val = qq{"$val"} ;
+				return $val ;
+			}),
+			SUB_FIX_MAKE_QUOTES	=>	sub {
+				my $val = shift ;
+				$val = qq{"$val"} ;
+				return $val ;
+			},
 		},
 		cygwin => {
 			ENV_VAR_PATH_SEP_CP	=>	';',
@@ -190,7 +241,3 @@ sub portable {
 
 
 1 ;
-
-
-
-
